"""Images subcommand implementation."""

from __future__ import annotations

import curses
import json
import shlex

from copy import deepcopy
from functools import partial
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.configuration_subsystem import Constants
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.image_manager import inspect_all
from ansible_navigator.runner import Command
from ansible_navigator.steps import Step
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import nonblocking_notification
from ansible_navigator.ui_framework import warning_notification
from ansible_navigator.utils.print import print_to_stdout

from . import _actions as actions
from . import run_action


if TYPE_CHECKING:
    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration


def filter_content_keys(obj: dict[Any, Any]) -> dict[Any, Any]:
    """Filter out some keys when showing image content.

    Args:
        obj: The object from which keys should be removed

    Returns:
        The object with keys removed
    """
    if isinstance(obj, list):
        working = [filter_content_keys(x) for x in obj]
        return working
    if isinstance(obj, dict):
        working = {}
        for k, val in obj.items():
            if not k.startswith("__"):
                working[k] = filter_content_keys(val)
        return working
    return obj


@actions.register
class Action(ActionBase):
    """Images subcommand implementation."""

    KEGEX = r"^im(?:ages)?(\s(?P<params>.*))?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:images`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="images")
        self._image_list: list[str] = []
        self._images = Step(
            name="images",
            step_type="menu",
            columns=["__image", "tag", "execution_environment", "created", "size"],
            value=[],
            select_func=self._build_image_menu,
        )

    def color_menu(self, colno: int, colname: str, entry: dict[str, Any]) -> tuple[int, int]:
        """Provide a color for a images menu entry in one column.

        Args:
            colno: The column number
            colname: The column name
            entry: The menu entry

        Returns:
            The color and decoration
        """
        if "__full_name" in entry:
            if entry.get("execution_environment") is False:
                return 8, 0
            if entry["__full_name"] == self._args.execution_environment_image:
                return 12, 0
            return 2, 0

        if self._images.selected:
            if self._images.selected["execution_environment"] is False:
                return 8, 0
            if self._images.selected["__full_name"] == self._args.execution_environment_image:
                return 12, 0
        return 2, 0

    def generate_content_heading(
        self,
        obj: dict[Any, Any],
        screen_w: int,
        name: str = "",
    ) -> CursesLines:
        """Create a heading for image content.

        Args:
            obj: The content going to be shown
            screen_w: The current screen width
            name: The name of the images menu

        Returns:
            The heading
        """
        text = ""
        if name == "image_menu":
            text = (
                f"{self.steps.previous.selected['first_column']}"
                f" ({self.steps.previous.selected['description']})"
            )
        elif name in ["python_package_list", "system_package_list"]:
            text = f"Name: {obj['name']} ({obj['version']})"

        color = 2
        if self._images.selected:
            if self._images.selected["__full_name"] == self._args.execution_environment_image:
                color = 4
            elif self._images.selected["execution_environment"] is False:
                color = 8

        empty_str = " " * (screen_w - len(text) + 1)
        heading_str = text + empty_str
        line_part = CursesLinePart(
            column=0,
            string=heading_str,
            color=color,
            decoration=curses.A_UNDERLINE,
        )
        return CursesLines((CursesLine((line_part,)),))

    def run_stdout(self) -> RunStdoutReturn:
        """Handle settings in mode stdout.

        Returns:
            RunStdoutReturn
        """
        self._logger.debug("images requested in stdout mode")

        details_source = self._args.entry("images_details").value.source
        if details_source is not Constants.DEFAULT_CFG:
            return self.run_stdout_details()

        self._collect_image_list()
        if not self._images.value:
            msg = "No images were found, or the configured container engine was not available."
            return RunStdoutReturn(message=msg, return_code=1)
        filtered = []
        for image in self._images.value:
            if image["execution_environment"] is False:
                continue
            image["name_tag"] = image["__name_tag"]
            image["full_name"] = image["__full_name"]
            filtered.append(filter_content_keys(image))
        print_to_stdout(
            content=filtered,
            content_format=getattr(ContentFormat, self._args.format.upper()),
            use_color=self._args.display_color,
        )
        return RunStdoutReturn(message="", return_code=0)

    def run_stdout_details(self) -> RunStdoutReturn:
        """Execute the ``images --details`` request for mode stdout.

        Returns:
            A message and return code
        """
        image_name = self._args.execution_environment_image
        sections: list[str] = self._args.entry("images_details").value.current

        output, error, return_code = self._run_runner(
            image_name=image_name,
            sections=sections,
        )
        if error or return_code:
            return RunStdoutReturn(message=error, return_code=return_code)

        details = self._parse(output)
        if details is None:
            message = "Image introspection failed, please check the logs and log an issue."
            return RunStdoutReturn(message=message, return_code=1)

        details.pop("errors")
        sections = self._args.entry("images_details").value.current
        for section_name, section in deepcopy(details).items():
            if section_name not in sections and "everything" not in sections:
                details.pop(section_name)
            else:
                for key in section:
                    if key.startswith("__"):
                        details[section_name].pop(key)

        details["image_name"] = image_name
        print_to_stdout(
            content=details,
            content_format=getattr(ContentFormat, self._args.format.upper()),
            use_color=self._args.display_color,
        )
        return RunStdoutReturn(message="", return_code=0)

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the ``images`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("images requested")
        self._prepare_to_run(app, interaction)

        args_updated = self._update_args(
            [self._name] + shlex.split(self._interaction.action.match.groupdict()["params"] or ""),
        )
        if not args_updated:
            self._prepare_to_exit(interaction)
            return None

        notification = nonblocking_notification(
            messages=["Collecting available images, this may take a minute..."],
        )
        interaction.ui.show_form(notification)

        self._collect_image_list()
        if not self._images.value:
            messages = [
                "No images were found, or the configured container engine was not available.",
            ]
            messages.append("Please check the log (:log) for errors.")
            warning = warning_notification(messages=messages)
            interaction.ui.show_form(warning)
            self._logger.error(messages[0])
            return None

        self.steps.append(self._images)

        while True:
            self._calling_app.update()
            self._take_step()

            if not self.steps:
                break

            if self.steps.current.name == "quit":
                return self.steps.current

        self._prepare_to_exit(interaction)
        return None

    def _take_step(self) -> None:
        """Take a step based on the current step or step back."""
        result = None
        if isinstance(self.steps.current, Interaction):
            result = run_action(self.steps.current.name, self.app, self.steps.current)
        elif isinstance(self.steps.current, Step):
            if self.steps.current.show_func:
                current_index = self.steps.current.index
                self.steps.current.show_func()
                self.steps.current.index = current_index

            if self.steps.current.type == "menu":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    columns=self.steps.current.columns,
                    color_menu_item=self.color_menu,
                )
            elif self.steps.current.type == "content":
                content_heading = partial(
                    self.generate_content_heading,
                    name=self.steps.previous.name,
                )
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    index=self.steps.current.index,
                    content_heading=content_heading,
                    filter_content_keys=filter_content_keys,
                )

        if result is None:
            self.steps.back_one()
        else:
            self.steps.append(result)

    def _build_image_content(self) -> Step:
        """Build the menu of image details.

        Returns:
            The image details menu definition
        """
        if self._images.selected is None:
            # an image should always be selected by now
            return self.steps.previous

        if self.steps.current.index == 0:
            step = Step(
                name="image_inspection",
                step_type="content",
                value=self._images.selected["inspect"],
            )
            return step

        if not self._images.selected["__introspected"]:
            message = "Collecting image details, this may take a minute..."
            notification = nonblocking_notification(messages=[message])
            self._interaction.ui.show_form(notification)
            introspection_success = self._introspect_image()
            if introspection_success is False:
                return self.steps.previous

        if self.steps.current.index == 1:
            step = Step(
                name="general",
                step_type="content",
                value=self._images.selected["general"],
            )

        elif self.steps.current.index == 2:
            step = Step(
                name="ansible",
                step_type="content",
                value=self._images.selected["ansible"],
            )

        elif self.steps.current.index == 3:
            step = Step(
                columns=["name", "version", "summary"],
                name="python_package_list",
                step_type="menu",
                select_func=self._build_python_content,
                value=sorted(
                    self._images.selected["python"]["details"],
                    key=lambda i: i["name"].lower(),
                ),
            )

        elif self.steps.current.index == 4:
            step = Step(
                columns=["name", "version", "summary"],
                name="system_package_list",
                step_type="menu",
                select_func=self._build_system_content,
                value=sorted(
                    self._images.selected["system"]["details"],
                    key=lambda i: i["name"].lower(),
                ),
            )

        elif self.steps.current.index == 5:
            step = Step(
                name="everything",
                step_type="content",
                value=self.steps.previous.selected,
            )
        return step

    def _build_image_menu(self) -> Step:
        """Build the menu of images.

        Returns:
            The images menu definition
        """
        if self._images.selected is None:
            # an image should always be selected by now
            return self.steps.previous

        first_column = f"Image: {self._images.selected['__name_tag']}"
        menu = [
            {
                first_column: "Image information",
                "description": "Information collected from image inspection",
            },
        ]
        if self.steps.current.selected["execution_environment"]:
            menu.append(
                {
                    first_column: "General information",
                    "description": "OS and python version information",
                },
            )
            menu.append(
                {
                    first_column: "Ansible version and collections",
                    "description": "Information about ansible and ansible collections",
                },
            )
            menu.append(
                {
                    first_column: "Python packages",
                    "description": "Information about python and python packages",
                },
            )
            menu.append(
                {
                    first_column: "Operating system packages",
                    "description": "Information about operating system packages",
                },
            )
            menu.append({first_column: "Everything", "description": "All image information"})
        for menu_entry in menu:
            menu_entry["first_column"] = first_column

        columns = [first_column, "description"]
        step = Step(
            columns=columns,
            name="image_menu",
            step_type="menu",
            value=menu,
            select_func=self._build_image_content,
        )
        self._interaction.ui.menu_filter(value=None)
        return step

    def _build_python_content(self) -> Step:
        """Build the content for an image's python packages.

        Returns:
            The python package content
        """
        return Step(
            name="python_content",
            step_type="content",
            value=self.steps.current.value,
            index=self.steps.current.index,
        )

    def _build_system_content(self) -> Step:
        """Build the content for an image's system packages.

        Returns:
            The system package content
        """
        return Step(
            name="system_content",
            step_type="content",
            value=self.steps.current.value,
            index=self.steps.current.index,
        )

    def _collect_image_list(self) -> None:
        """Build a list of images for the images menu."""
        images, error = inspect_all(container_engine=self._args.container_engine)
        if error or not images:
            self._logger.error(error)
            return

        for image in images:
            image["__introspected"] = False
            image["name"] = image["repository"].split("/")[-1]
            image["__image"] = image["name"]
            image["__name_tag"] = f"{image['name']}:{image['tag']}"
            image["__full_name"] = f"{image['repository']}:{image['tag']}"
            image["__name"] = image["name"]
            if image["__full_name"] == self._args.execution_environment_image:
                image["__name"] += " (primary)"
                image["__name_tag"] += " (primary)"

            details = image["inspect"]["details"]

            try:
                legacy_check = details["config"]["working_dir"] == "/runner"
            except (KeyError, TypeError):
                legacy_check = False

            # podman has a root label
            try:
                root_label_check = details["labels"]["ansible-execution-environment"] == "true"
            except (KeyError, TypeError):
                root_label_check = False

            # docker has only a config.label
            try:
                config_label_check = (
                    details["config"]["labels"]["ansible-execution-environment"] == "true"
                )
            except (KeyError, TypeError):
                config_label_check = False

            image["execution_environment"] = any(
                (legacy_check, root_label_check, config_label_check),
            )
        self._images.value = sorted(images, key=lambda i: i["name"])

    def _introspect_image(self) -> bool:
        """Use the runner subsystem to introspect an image.

        Returns:
            An indication of image introspection success
        """
        if self._images.selected is None:
            # an image should always be selected by now
            return False

        self._images.selected["__introspected"] = True

        output, error, _return_code = self._run_runner(
            image_name=self._images.selected["__full_name"],
        )

        if error:
            self._logger.error(
                "Image introspection failed (runner), the return value was: %s",
                error,
            )
            self.notify_failed()
            return False
        parsed = self._parse(output)
        if parsed is None:
            self.notify_failed()
            return False

        try:
            self._images.selected["general"] = {
                "os": parsed["os_release"],
                "friendly": parsed["redhat_release"],
                "python": parsed["python_version"],
            }
            self._images.selected["ansible"] = {
                "ansible": {
                    "collections": parsed["ansible_collections"],
                    "version": parsed["ansible_version"],
                },
            }
            self._images.selected["python"] = parsed["python_packages"]
            self._images.selected["system"] = parsed["system_packages"]
        except KeyError:
            self._logger.exception(
                "Image introspection failed (keys), the return value was: %s",
                output[0:1000],
            )
            self.notify_failed()
            return False
        return True

    def _parse(self, output: str) -> dict[Any, Any] | None:
        """Load and process the ``json`` output from the image introspection process.

        Args:
            output: The output from the image introspection process

        Returns:
            The parsed output
        """
        try:
            if not output.startswith("{"):
                _warnings, json_str = output.split("{", 1)
                json_str = "{" + json_str
            else:
                json_str = output
            parsed = json.loads(json_str)
            self._logger.debug("json loading output succeeded")
        except (json.decoder.JSONDecodeError, ValueError) as exc:
            self._logger.exception("Unable to extract introspection from stdout")
            self._logger.debug("error json loading output: '%s'", str(exc))
            self._logger.debug(output)
            self._logger.exception(
                "Image introspection failed (parsed), the return value was: %s",
                output[0:1000],
            )
            return None

        for error in parsed["errors"]:
            self._logger.error("%s %s", error["path"], error["error"])
        return parsed

    def _run_runner(
        self,
        image_name: str,
        sections: list[str] | None = None,
    ) -> tuple[str, str, int]:
        """Run runner to collect image details.

        Args:
            image_name: The full image name
            sections: Optional list of introspection sections to collect.
                When provided, only the named collectors run inside the
                container instead of all of them.

        Returns:
            Output, errors and the return code
        """
        cache_path = self._args.internals.cache_path
        container_volume_mounts = [f"{cache_path}:{cache_path}"]
        python_exec_path = f"{cache_path}/python_latest.sh"

        cmdline = [f"{cache_path}/image_introspect.py"]
        if sections:
            cmdline.extend(["--sections", *sections])

        kwargs = {
            "cmdline": cmdline,
            "container_engine": self._args.container_engine,
            "container_volume_mounts": container_volume_mounts,
            "execution_environment_image": image_name,
            "execution_environment": True,
            "navigator_mode": "interactive",
        }

        if isinstance(self._args.container_options, list):
            kwargs.update({"container_options": self._args.container_options})

        self._logger.debug(
            "Invoke runner with executable_cmd: %s and kwargs: %s",
            python_exec_path,
            kwargs,
        )
        _runner = Command(executable_cmd=python_exec_path, **kwargs)
        output, error, return_code = _runner.run()
        return output, error, return_code

    def notify_failed(self) -> None:
        """Notify image introspection failed."""
        msgs = ["humph. Something went really wrong while introspecting the image."]
        msgs.append("Details have been added to the log file")
        closing = ["[HINT] Please log an issue about this one, it shouldn't have happened"]
        warning = warning_notification(messages=msgs + closing)
        self._interaction.ui.show_form(warning)



"""Menu selection implementation.

Processor of a menu selection from a numeric key press
or entry at the single line prompt. e.g ``:42``
"""

import logging

from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


@actions.register
class Action:
    """Menu selection implementation."""

    KEGEX = r"^\d+$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the select action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def run(self, interaction: Interaction, app: AppPublic) -> None:
        """Execute a menu selection for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance
        """
        self._logger.debug("selection made")
        interaction.ui.scroll(0)
        interaction.ui.clear()
        this = app.steps.back_one()  # remove this
        app.steps.current.index = interaction.action.value  # update index
        app.steps.append(app.steps.current.select_func())  # add next
        app.steps.append(this)  # put this back on stack
        self._logger.debug(
            "Requested next step in %s will be %s",
            app.name,
            app.steps.previous.name,
        )



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import NamedTuple


if TYPE_CHECKING:
    from .rules import Entry
    from .rules import WhileRule


class State(NamedTuple):
    entries: tuple[Entry, ...]
    while_stack: tuple[tuple[WhileRule, int], ...]

    @classmethod
    def root(cls, entry: Entry) -> State:
        return cls((entry,), ())

    @property
    def cur(self) -> Entry:
        return self.entries[-1]

    def push(self, entry: Entry) -> State:
        return self._replace(entries=(*self.entries, entry))

    def pop(self) -> State:
        return self._replace(entries=self.entries[:-1])

    def push_while(self, rule: WhileRule, entry: Entry) -> State:
        entries = (*self.entries, entry)
        while_stack = (*self.while_stack, (rule, len(entries)))
        return self._replace(entries=entries, while_stack=while_stack)

    def pop_while(self) -> State:
        entries, while_stack = self.entries[:-1], self.while_stack[:-1]
        return self._replace(entries=entries, while_stack=while_stack)



"""Template command implementation.

Processor of a template request at the single line prompt. e.g. {{ }}
"""

from __future__ import annotations

import html

from collections.abc import Mapping
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import warning_notification
from ansible_navigator.utils.functions import remove_dbl_un
from ansible_navigator.utils.functions import templar

from . import _actions as actions


if TYPE_CHECKING:
    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration


@actions.register
class Action(ActionBase):
    """Template command implementation."""

    KEGEX = r"^{{\s*(?P<params>.*?)\s*}}$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the template action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="template")

    def _resolve_template_vars(
        self,
        interaction: Interaction,
    ) -> tuple[dict[str, Any], list[str]] | None:
        """Resolve template variables from the current interaction content or menu.

        Args:
            interaction: The interaction from the user

        Returns:
            A tuple of (template_vars, type_msgs) or None if no content is available
        """
        type_msgs = ["Current content passed for templating is not a dictionary."]
        type_msgs.append("[HINT] Use 'this' to reference it (e.g. {{ this[0] }}")

        if interaction.content:
            if isinstance(interaction.content.showing, Mapping):
                return dict(interaction.content.showing), []
            return {"this": interaction.content.showing}, type_msgs

        if interaction.menu:
            obj = [
                {remove_dbl_un(k): v for k, v in c.items() if k in interaction.menu.columns}
                for c in interaction.menu.current
            ]
            if interaction.ui.menu_filter():
                obj = [
                    e
                    for e in obj
                    if interaction.ui.menu_filter().search(" ".join(str(v) for v in e.values()))
                ]
            return {"this": obj}, type_msgs

        self._logger.error("No menu or content found")
        return None

    @staticmethod
    def _determine_content_format(
        requested: str,
        templated: Any,
    ) -> ContentFormat | None:
        """Determine the content format based on the requested parameter and templated value.

        Args:
            requested: The requested template parameter string
            templated: The templated output value

        Returns:
            The content format or None
        """
        if requested == "examples":
            return ContentFormat.YAML_TXT
        if requested == "readme":
            return ContentFormat.MARKDOWN
        if isinstance(templated, str):
            return ContentFormat.TXT
        return None

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the templating request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("template requested '%s'", interaction.action.value)
        self._prepare_to_run(app, interaction)

        resolved = self._resolve_template_vars(interaction)
        if resolved is None:
            return None
        template_vars, type_msgs = resolved

        previous_scroll = interaction.ui.scroll()
        interaction.ui.scroll(0)

        errors, templated = templar(
            string=str(interaction.action.value),
            template_vars=template_vars,
        )
        if errors:
            msgs = ["Errors encountered while templating input"] + errors
            msgs.extend(type_msgs)
            interaction.ui.show_form(warning_notification(msgs))
            return None

        if isinstance(templated, str):
            templated = html.unescape(templated)

        requested = self._interaction.action.match.groupdict()["params"]
        content_format = self._determine_content_format(requested, templated)

        while True:
            app.update()
            next_interaction: Interaction = interaction.ui.show(
                obj=templated,
                content_format=content_format,
            )
            if next_interaction.name != "refresh":
                break

        interaction.ui.scroll(previous_scroll)
        self._prepare_to_exit(interaction)
        return next_interaction



"""Doc subcommand implementation."""

from __future__ import annotations

import curses
import json
import shlex
import shutil

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.configuration_subsystem import Constants as C
from ansible_navigator.runner import AnsibleDoc
from ansible_navigator.runner import Command
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


if TYPE_CHECKING:
    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration


@actions.register
class Action(ActionBase):
    """Doc subcommand implementation."""

    KEGEX = r"^d(?:oc)?(\s(?P<params>.*))?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``doc`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="doc")

        self._plugin_name: str | None = None
        self._plugin_type: str | None = None
        self._runner: Command | AnsibleDoc

    def generate_content_heading(self, _obj: dict[Any, Any], screen_w: int) -> CursesLines:
        """Create a heading for doc content.

        Args:
            _obj: The content going to be shown
            screen_w: The current screen width

        Returns:
            The heading
        """
        plugin_str = f"Name: {self._plugin_name} ({self._plugin_type})"
        empty_str = " " * (screen_w - len(plugin_str) + 1)
        heading_str = (plugin_str + empty_str).capitalize()

        line_part = CursesLinePart(
            column=0,
            string=heading_str,
            color=0,
            decoration=curses.A_UNDERLINE,
        )

        return CursesLines((CursesLine((line_part,)),))

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the ``doc`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("doc requested in interactive")
        self._prepare_to_run(app, interaction)

        colon_prompt = self._interaction.action.match.groupdict()["params"]
        source = None

        # Nothing provided at colon prompt and content is showing, get the task action
        if interaction.content and not colon_prompt:
            try:
                self._plugin_name = interaction.content.showing["task_action"]
                self._plugin_type = self._args.entry("plugin_type").value.default
                source = "task action"

            except (KeyError, AttributeError, TypeError):
                self._logger.info("No plugin name found in current content")

        # Process the colon prompt and allow update args to identify missing entries
        if self._plugin_name is None:
            args_updated = self._update_args([self._name] + shlex.split(colon_prompt or ""))
            if not args_updated:
                self._prepare_to_exit(interaction)
                return None
            source = self._args.entry("plugin_name").value.source.value
            self._plugin_name = self._args.plugin_name
            self._plugin_type = self._args.plugin_type

        self._logger.debug("Plugin name used from %s: %s", source, self._plugin_name)
        self._logger.debug("Plugin type used from %s: %s", source, self._plugin_type)

        plugin_doc = self._run_runner()

        if not isinstance(plugin_doc, dict):
            self._prepare_to_exit(interaction)
            return None

        while True:
            app.update()
            next_interaction: Interaction = interaction.ui.show(
                content_heading=self.generate_content_heading,
                obj=plugin_doc,
            )
            if next_interaction.name != "refresh":
                break

        self._prepare_to_exit(interaction)
        return next_interaction

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``doc`` request for mode stdout.

        Returns:
            The return code or 1. If the response from the runner
            invocation is None, indicates there is no console output to
            display, so assume an issue and return 1 along with a
            message to review the logs.
        """
        self._plugin_name = self._args.plugin_name
        self._plugin_type = self._args.plugin_type
        self._logger.debug("doc requested in stdout mode")
        response = self._run_runner()
        if response is None:
            self._logger.error("Unexpected response: %s", response)
            return RunStdoutReturn(message="Please review the log for errors.", return_code=1)
        _out, error, return_code = response
        return RunStdoutReturn(message=error, return_code=return_code)

    def _run_runner(self) -> dict[Any, Any] | tuple[str, str, int] | None:
        """Use the runner subsystem to retrieve the configuration.

        Returns:
            For mode interactive nothing or the plugin's doc. For mode
            stdout the output, errors and return code from runner.
        """
        if isinstance(self._args.set_environment_variable, dict):
            set_env_vars = {**self._args.set_environment_variable}
        else:
            set_env_vars = {}

        if self._args.display_color is False or self._args.mode == "interactive":
            set_env_vars["ANSIBLE_NOCOLOR"] = "1"

        kwargs = {
            "container_engine": self._args.container_engine,
            "execution_environment_image": self._args.execution_environment_image,
            "execution_environment": self._args.execution_environment,
            "navigator_mode": self._args.mode,
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": set_env_vars,
            "private_data_dir": self._args.ansible_runner_artifact_dir,
            "rotate_artifacts": self._args.ansible_runner_rotate_artifacts_count,
            "timeout": self._args.ansible_runner_timeout,
        }

        if isinstance(self._args.execution_environment_volume_mounts, list):
            kwargs.update(
                {"container_volume_mounts": self._args.execution_environment_volume_mounts},
            )

        if isinstance(self._args.container_options, list):
            kwargs.update({"container_options": self._args.container_options})

        if self._args.mode == "interactive":
            return self._run_runner_interactive(kwargs)
        return self._run_runner_stdout(kwargs)

    def _run_runner_interactive(
        self,
        kwargs: dict[str, Any],
    ) -> dict[Any, Any] | None:
        """Run the ansible-doc runner in interactive mode.

        Resolves the playbook directory, fetches plugin documentation
        via AnsibleDoc, and extracts the result.

        Args:
            kwargs: The common runner keyword arguments.

        Returns:
            The plugin's documentation dictionary or None.
        """
        if isinstance(self._args.playbook, str):
            playbook_dir = f"{Path(self._args.playbook).parent}"
        else:
            playbook_dir = str(Path.cwd())
        kwargs.update({"host_cwd": playbook_dir})

        self._runner = AnsibleDoc(**kwargs)

        # set the playbook directory so playbook
        # adjacent collection docs can be found
        self._logger.debug("doc playbook dir set to: %s", playbook_dir)

        plugin_doc, plugin_doc_err = self._runner.fetch_plugin_doc(
            [self._plugin_name],
            plugin_type=self._plugin_type,
            playbook_dir=playbook_dir,
        )
        if plugin_doc_err:
            self._logger.error(
                "Error occurred while fetching doc for plugin %s: '%s'",
                self._plugin_name,
                plugin_doc_err,
            )
        plugin_doc_response = self._extract_plugin_doc(plugin_doc, plugin_doc_err)
        return plugin_doc_response

    def _run_runner_stdout(
        self,
        kwargs: dict[str, Any],
    ) -> tuple[str, str, int] | None:
        """Run the ansible-doc runner in stdout mode.

        Resolves the ansible-doc executable path, builds the
        pass-through arguments, and runs the command.

        Args:
            kwargs: The common runner keyword arguments.

        Raises:
            RuntimeError: When the ansible-doc command cannot be
                found with execution environment support disabled.

        Returns:
            The output, errors and return code from runner.
        """
        kwargs.update({"host_cwd": str(Path.cwd())})
        if self._args.execution_environment:
            ansible_doc_path = "ansible-doc"
        else:
            exec_path = shutil.which("ansible-doc")
            if exec_path is None:
                msg = "'ansible-doc' executable not found"
                self._logger.error(msg)
                raise RuntimeError(msg)
            ansible_doc_path = exec_path

        pass_through_arg = []
        if self._plugin_name is not C.NOT_SET:
            pass_through_arg.append(self._plugin_name)

        if self._plugin_type is not C.NOT_SET:
            pass_through_arg.extend(["-t", self._plugin_type])

        if self._args.help_doc is True:
            pass_through_arg.append("--help")

        if isinstance(self._args.cmdline, list):
            pass_through_arg.extend(self._args.cmdline)

        kwargs.update({"cmdline": pass_through_arg})

        self._runner = Command(executable_cmd=ansible_doc_path, **kwargs)
        stdout_return = self._runner.run()
        return stdout_return

    def _extract_plugin_doc(
        self,
        out: dict[Any, Any] | str,
        err: dict[Any, Any] | str,
    ) -> dict[Any, Any] | None:
        """Extract the plugin's documentation from the runner output.

        Args:
            out: The output from runner
            err: Any runner errors

        Returns:
            The plugin's doc or errors
        """
        if self._args.execution_environment:
            error_key_name = "execution_environment_errors"
        else:
            error_key_name = "local_errors"

        if out:
            return self._process_plugin_output(out, err, error_key_name)

        plugin_doc: dict[Any, Any] = {}
        if err:
            if isinstance(err, dict):
                plugin_doc[error_key_name] = err
            else:
                plugin_doc[error_key_name] = err.splitlines()
        return plugin_doc

    def _process_plugin_output(
        self,
        out: dict[Any, Any] | str,
        err: dict[Any, Any] | str,
        error_key_name: str,
    ) -> dict[Any, Any]:
        """Process the plugin output when output is present.

        Parses the runner output (dict or JSON string), extracts the
        plugin documentation, and attaches any warnings from stderr.

        Args:
            out: The non-empty output from runner.
            err: Any runner errors.
            error_key_name: The key name to use for error entries.

        Returns:
            The plugin's documentation dictionary with warnings.
        """
        plugin_doc: dict[Any, Any] = {}
        if isinstance(out, dict):
            plugin_doc = out[self._plugin_name]
        else:
            try:
                json_loaded = json.loads(out)
            except json.JSONDecodeError as exc:
                if self._args.mode == "interactive":
                    self._logger.info(
                        "Parsing of ansible-doc output failed for '%s'",
                        self._plugin_name,
                    )
                self._logger.debug("json decode error: %s", str(exc))
                self._logger.debug("tried: %s", out)
                plugin_doc[error_key_name] = out
            else:
                plugin_doc = json_loaded[self._plugin_name]

        if isinstance(err, dict):
            plugin_doc["warnings"] = err
        else:
            plugin_doc["warnings"] = err.splitlines()

        return plugin_doc



"""The ability to perform migrations."""

from __future__ import annotations

from pathlib import Path
from typing import Any

from ansible_navigator.utils.ansi import COLOR
from ansible_navigator.utils.ansi import IS_TTY
from ansible_navigator.utils.ansi import blank_line
from ansible_navigator.utils.ansi import info
from ansible_navigator.utils.ansi import prompt_enter
from ansible_navigator.utils.ansi import prompt_yn
from ansible_navigator.utils.ansi import success
from ansible_navigator.utils.ansi import warning

from .definitions import MigrationType
from .definitions import migrations


# isort: off
# pylint: disable=unused-import
# Migrations will run in the order they are imported
from .v1_v2_settings_file import V1V2SettingsFile  # noqa: F401

# isort: on


def run_all_migrations(migration_types: tuple[MigrationType], settings_file_str: str = "") -> None:
    """Run all migrations.

    Args:
        migration_types: Type of migration
        settings_file_str: Path to the settings file
    """
    for migration_type in migration_types:
        run_migrations(settings_file_str, migration_type)


def _collect_migrations(
    settings_file_str: str,
    migration_type: MigrationType,
) -> list[Any]:
    """Instantiate, configure, and check-run migrations of a given type.

    Args:
        settings_file_str: Path to the settings file.
        migration_type: Type of migration.

    Returns:
        The list of configured migration instances.
    """
    migrations_to_run = [
        migration()
        for migration in migrations
        if getattr(migration, "migration_type", None) == migration_type
    ]
    for migration in migrations_to_run:
        migration.check = True
        if getattr(migration, "migration_type", None) == MigrationType.SETTINGS_FILE:
            migration.settings_file_path = Path(settings_file_str)
        migration.run()
    return migrations_to_run


def _display_needed_migrations(migrations_to_run: list[Any]) -> None:
    """Display the list of needed migrations.

    Args:
        migrations_to_run: The list of migration instances.
    """
    tense = "are" if IS_TTY else "were"
    blank_line()
    warning(color=COLOR, message=f"The following version migrations {tense} required:")
    for migration in migrations_to_run:
        if migration.needed_now:
            info(color=COLOR, message=f"  - {migration.name}")


def _execute_migrations(migrations_to_run: list[Any]) -> None:
    """Run migrations and report success or failure.

    Args:
        migrations_to_run: The list of migration instances.
    """
    for migration in migrations_to_run:
        blank_line()
        info(color=COLOR, message=f"{migration.name}:")
        migration.check = False
        migration.run()
    if not any(migration.needed_now for migration in migrations_to_run):
        blank_line()
        success(color=COLOR, message="Migration complete")
    else:
        blank_line()
        warning(
            color=COLOR,
            message="The following version migrations could not be completed:",
        )
        for migration in migrations_to_run:
            if migration.needed_now:
                info(color=COLOR, message=f" - {migration.name}")


def run_migrations(settings_file_str: str, migration_type: MigrationType) -> None:
    """Run migrations of a specific type.

    Args:
        settings_file_str: Path to the settings file
        migration_type: Type of migration
    """
    migrations_to_run = _collect_migrations(settings_file_str, migration_type)
    migratable = [migration.name for migration in migrations_to_run if migration.needed_now]

    if not migratable:
        return

    _display_needed_migrations(migrations_to_run)

    if not IS_TTY:
        blank_line()
        warning(color=COLOR, message="Migration not possible without a TTY")
        return

    if prompt_yn(message="Do you want to run them all?"):
        _execute_migrations(migrations_to_run)
    else:
        blank_line()
        warning(color=COLOR, message="Migration cancelled")
    blank_line()
    prompt_enter()



"""Herein lies the base class for running commands using ansible-runner.

All attributes common to a subprocess or async command are defined here.
"""

from __future__ import annotations

import sys

from typing import Any

from .base import Base


class CommandBase(Base):
    """Base class for runner command interaction."""

    def __init__(
        self,
        executable_cmd: str,
        cmdline: list[str] | None = None,
        playbook: str | None = None,
        inventory: list[str] | None = None,
        **kwargs: Any,
    ) -> None:
        """Handle common arguments of ``run_command`` interface for ``ansible-runner``.

        Args:
            executable_cmd: The command to be invoked
            cmdline: A list of arguments to be passed to the executable
                command
            playbook: The playbook file name to run
            inventory: List of path to the inventory files
            **kwargs: The arguments for the runner call
        """
        self._executable_cmd = executable_cmd
        self._cmdline: list[str] = cmdline if isinstance(cmdline, list) else []
        self._playbook = playbook
        self._inventory: list[str] = inventory if isinstance(inventory, list) else []
        super().__init__(**kwargs)

    def generate_run_command_args(self) -> None:
        """Generate arguments required to be passed to ansible-runner."""
        if self._playbook:
            self._cmdline.insert(0, self._playbook)

        for inv in self._inventory:
            self._cmdline.extend(["-i", inv])

        self._runner_args["executable_cmd"] = self._executable_cmd
        self._runner_args["cmdline_args"] = self._cmdline

        if self._navigator_mode == "stdout":
            self._runner_args.update(
                {"input_fd": sys.stdin, "output_fd": sys.stdout, "error_fd": sys.stderr},
            )

        for key, value in self._runner_args.items():
            self._logger.debug("Runner arg: %s:%s", key, value)
